#ifndef __COMPUTATION_GRAPH_H__
#define __COMPUTATION_GRAPH_H__

#include <stdio.h>
#include <math.h>

typedef void (*execute_node_t)(void *state,
                               const double* params,
                               const double *inputs,
                               double *outputs);

typedef void (*reset_node_t)(void *state);

enum node_processed_state {
    UNPROCESSED,
    DISCOVERED,
    PROCESSED
};

struct computation_graph {
    int n_nodes;
    int size;
    struct graph_node *nodes;
};

// Declares a node type
struct graph_node_type {
    const char* const* input_names;
    const char* const* output_names;
    const char* const* param_names;
    int n_inputs;
    int n_outputs;
    int n_params;
    execute_node_t execute;
    reset_node_t reset;
};

// Declares an instance of a node
struct graph_node {
    const char *name;
    const struct graph_node_type* type;
    double *output_values;
    double *param_values;
    int n_children;
    void *state;
    int processed_state;
    struct input_type {
        int controller_id;
        int controller_output;
    } *input_srcs;
};

/*
 * Creates an empty computation graph
 * May return NULL on failure
 */
struct computation_graph *create_graph();

/*
 * Defines which node's output gets its value passed into the input of a different node.
 * Will call reset for each node which was previously orphaned, but is now connected to the graph
 * dest_node_id: The ID of the node to which the input belongs
 * dest_input: The ID of the input for node dest_cntl to pass the value into
 * src_node_id: The node ID where the value is coming from
 * src_output: The ID of the output on <src_node_id> where the value comes from
 * Returns 0 for success
 */
int graph_set_source(struct computation_graph *graph, int dest_node_id, int dest_input, int src_node_id, int src_output);

/*
 * Creates a new node with the given data, and adds it to the graph.
 * Returns a negative integer upon failure.
 * Otherwise returns a positive integer which is the ID of the new node
 */
int graph_add_node(struct computation_graph *graph,
                   const char *name,
                   const struct graph_node_type *type,
                   void *state);

/*
 * Returns the value at the output of the requested node for the requested output.
 * Returns 0 if the given node or output IDs are invalid
 */
double graph_get_output(const struct computation_graph *graph, int node_id, int output_id);

/*
 * Sets a parameter given by param_id on node node_id to the given value
 * Returns 0 upon success
 */
int graph_set_param_val(struct computation_graph *graph, int node_id, int param_id, double value);

/*
 * Returns the value of the param at param_id for the given node
 * Will return NaN if the given node or parameter IDs are not valid
 */
double graph_get_param_val(struct computation_graph *graph, int node_id, int param_id);

/*
 * Computes the nodes given by node_id.
 * To do so, computes all nodes which are ancestors of each given node_id in topological order, with
 * the final computation being node_id itself.
 */
void graph_compute_nodes(struct computation_graph *graph, int* node_ids, int n_nodes);

/*
 * Writes a graphical representation of the given graph to <of> in the DOT language
 */
int export_dot(const struct computation_graph* graph, FILE* of, int print_outputs);

#endif // __COMPUTATION_GRAPH_H__
