/*
 * control_algorithm.c
 *
 *  Created on: Feb 20, 2016
 *      Author: ucart
 */

// This implemented modular quadrotor software implements a PID control algorithm

#include "control_algorithm.h"
#include "graph_blocks/node_pid.h"
#include "graph_blocks/node_bounds.h"
#include "graph_blocks/node_constant.h"
#include "graph_blocks/node_mixer.h"
#include "PID.h"
#include "util.h"

#define ROLL_PITCH_MAX_ANGLE 0.3490 // 20 degrees

 int control_algorithm_init(parameter_t * ps)
 {
    struct computation_graph* graph = create_graph();
    ps->graph = graph;

    // Create all the PID blocks
    ps->roll_pid = graph_add_node_pid(graph, "Roll PID");
    ps->pitch_pid = graph_add_node_pid(graph, "Pitch PID");
    ps->yaw_pid = graph_add_node_pid(graph, "Yaw PID");
    ps->roll_r_pid = graph_add_node_pid(graph, "Roll Rate PID");
    ps->pitch_r_pid = graph_add_node_pid(graph, "Pitch Rate PID");
    ps->yaw_r_pid = graph_add_node_pid(graph, "Yaw Rate PID");
    ps->x_pos_pid = graph_add_node_pid(graph, "X pos PID");
    ps->y_pos_pid = graph_add_node_pid(graph, "Y pos PID");
    ps->alt_pid = graph_add_node_pid(graph, "Altitude");

    // Create blocks for bounds checking
    ps->clamp_d_pwmP = graph_add_node_bounds(graph, "P PWM Clamp");
    ps->clamp_d_pwmR = graph_add_node_bounds(graph, "R PWM Clamp");
    ps->clamp_d_pwmY = graph_add_node_bounds(graph, "Y PWM Clamp");
    ps->clamp_pitch = graph_add_node_bounds(graph, "Pitch Clamp");
    ps->clamp_roll= graph_add_node_bounds(graph, "Roll Clamp");

    // Create blocks for sensor inputs
    ps->cur_pitch = graph_add_node_const(graph, "Pitch");
    ps->cur_roll = graph_add_node_const(graph, "Roll");
    ps->cur_yaw = graph_add_node_const(graph, "Yaw");
	// Yaw angular velocity PID
    // theta_dot is the angular velocity about the y-axis
    // phi_dot is the angular velocity about the x-axis
	// psi_dot is the angular velocity about the z-axis
	// These are calculated from using the gimbal equations
    ps->theta_dot = graph_add_node_const(graph, "dTheta");
    ps->phi_dot = graph_add_node_const(graph, "dPhi");
    ps->psi_dot = graph_add_node_const(graph, "dPsi");

    // Create blocks for RC controller
    ps->rc_pitch = graph_add_node_const(graph, "RC Pitch");
    ps->rc_roll = graph_add_node_const(graph, "RC Roll");
    ps->rc_yaw = graph_add_node_const(graph, "RC Yaw");
    ps->rc_throttle = graph_add_node_const(graph, "RC Throttle");

    ps->mixer = graph_add_node_mixer(graph, "Signal Mixer");

    ps->dt = graph_add_node_const(graph, "dT");

    // Connect pitch PID chain
    graph_set_source(graph, ps->pitch_r_pid, PID_SETPOINT, ps->pitch_pid, PID_CORRECTION);
    graph_set_source(graph, ps->pitch_r_pid, PID_CUR_POINT, ps->theta_dot, CONST_VAL);
    graph_set_source(graph, ps->pitch_r_pid, PID_DT, ps->dt, CONST_VAL);
    graph_set_source(graph, ps->pitch_pid, PID_SETPOINT, ps->rc_pitch, CONST_VAL);
    graph_set_source(graph, ps->pitch_pid, PID_CUR_POINT, ps->cur_pitch, CONST_VAL);
    graph_set_source(graph, ps->pitch_pid, PID_DT, ps->dt, CONST_VAL);

     // Connect roll PID chain
    graph_set_source(graph, ps->roll_r_pid, PID_SETPOINT, ps->roll_pid, PID_CORRECTION);
    graph_set_source(graph, ps->roll_r_pid, PID_CUR_POINT, ps->phi_dot, CONST_VAL);
    graph_set_source(graph, ps->roll_r_pid, PID_DT, ps->dt, CONST_VAL);
    graph_set_source(graph, ps->roll_pid, PID_SETPOINT, ps->rc_roll, CONST_VAL);
    graph_set_source(graph, ps->roll_pid, PID_CUR_POINT, ps->cur_roll, CONST_VAL);
    graph_set_source(graph, ps->roll_pid, PID_DT, ps->dt, CONST_VAL);

    // Connect yaw PID chain
    graph_set_source(graph, ps->yaw_r_pid, PID_SETPOINT, ps->rc_yaw, PID_CORRECTION);
    graph_set_source(graph, ps->yaw_r_pid, PID_CUR_POINT, ps->psi_dot, CONST_VAL);
    graph_set_source(graph, ps->yaw_r_pid, PID_DT, ps->dt, CONST_VAL);
    /*
    graph_set_source(graph, ps->yaw_pid, PID_SETPOINT, ps->rc_yaw, CONST_VAL);
    graph_set_source(graph, ps->yaw_pid, PID_CUR_POINT, ps->yaw, CONST_VAL);
    graph_set_source(graph, ps->yaw_pid, PID_DT, ps->dt, CONST_VAL);
    */

    // Connect angle clamping blocks
    graph_set_source(graph, ps->clamp_pitch, BOUNDS_IN, ps->y_pos_pid, PID_CORRECTION);
    graph_set_source(graph, ps->clamp_roll, BOUNDS_IN, ps->x_pos_pid, PID_CORRECTION);

    // Connect PWM clamping blocks
    graph_set_source(graph, ps->clamp_d_pwmP, BOUNDS_IN, ps->pitch_r_pid, PID_CORRECTION);
    graph_set_source(graph, ps->clamp_d_pwmR, BOUNDS_IN, ps->roll_r_pid, PID_CORRECTION);
    graph_set_source(graph, ps->clamp_d_pwmY, BOUNDS_IN, ps->yaw_r_pid, PID_CORRECTION);

    // Connect signal mixer
    graph_set_source(graph, ps->mixer, MIXER_THROTTLE, ps->rc_throttle, CONST_VAL);
    graph_set_source(graph, ps->mixer, MIXER_PITCH, ps->clamp_d_pwmP, BOUNDS_OUT);
    graph_set_source(graph, ps->mixer, MIXER_ROLL, ps->clamp_d_pwmR, BOUNDS_OUT);
    graph_set_source(graph, ps->mixer, MIXER_YAW, ps->clamp_d_pwmY, BOUNDS_OUT);

    // Set pitch PID constants
    graph_set_param_val(graph, ps->pitch_pid, PID_KP, PITCH_ANGLE_KP);
    graph_set_param_val(graph, ps->pitch_pid, PID_KI, PITCH_ANGLE_KI);
    graph_set_param_val(graph, ps->pitch_pid, PID_KD, PITCH_ANGLE_KD);
    graph_set_param_val(graph, ps->pitch_r_pid, PID_KP, PITCH_ANGULAR_VELOCITY_KP);
    graph_set_param_val(graph, ps->pitch_r_pid, PID_KI, PITCH_ANGULAR_VELOCITY_KI);
    graph_set_param_val(graph, ps->pitch_r_pid, PID_KD, PITCH_ANGULAR_VELOCITY_KD);

    // Set roll PID constants
    graph_set_param_val(graph, ps->roll_pid, PID_KP, ROLL_ANGLE_KP);
    graph_set_param_val(graph, ps->roll_pid, PID_KI, ROLL_ANGLE_KI);
    graph_set_param_val(graph, ps->roll_pid, PID_KD, ROLL_ANGLE_KD);
    graph_set_param_val(graph, ps->roll_r_pid, PID_KP, ROLL_ANGULAR_VELOCITY_KP);
    graph_set_param_val(graph, ps->roll_r_pid, PID_KI, ROLL_ANGULAR_VELOCITY_KI);
    graph_set_param_val(graph, ps->roll_r_pid, PID_KD, ROLL_ANGULAR_VELOCITY_KD);

    // Set yaw PID constants
    graph_set_param_val(graph, ps->yaw_pid, PID_KP, YAW_ANGLE_KP);
    graph_set_param_val(graph, ps->yaw_pid, PID_KI, YAW_ANGLE_KI);
    graph_set_param_val(graph, ps->yaw_pid, PID_KD, YAW_ANGLE_KD);
    graph_set_param_val(graph, ps->yaw_r_pid, PID_KP, YAW_ANGULAR_VELOCITY_KP);
    graph_set_param_val(graph, ps->yaw_r_pid, PID_KI, YAW_ANGULAR_VELOCITY_KI);
    graph_set_param_val(graph, ps->yaw_r_pid, PID_KD, YAW_ANGULAR_VELOCITY_KD);

    // Set angle clamping limits
    graph_set_param_val(graph, ps->clamp_pitch, BOUNDS_MIN, -ROLL_PITCH_MAX_ANGLE);
    graph_set_param_val(graph, ps->clamp_pitch, BOUNDS_MAX, ROLL_PITCH_MAX_ANGLE);
    graph_set_param_val(graph, ps->clamp_roll, BOUNDS_MIN, -ROLL_PITCH_MAX_ANGLE);
    graph_set_param_val(graph, ps->clamp_roll, BOUNDS_MAX, ROLL_PITCH_MAX_ANGLE);

    // Set PWM difference clamping limits
    graph_set_param_val(graph, ps->clamp_d_pwmP, BOUNDS_MIN, -20000);
    graph_set_param_val(graph, ps->clamp_d_pwmP, BOUNDS_MAX, 20000);
    graph_set_param_val(graph, ps->clamp_d_pwmR, BOUNDS_MIN, -20000);
    graph_set_param_val(graph, ps->clamp_d_pwmR, BOUNDS_MAX, 20000);
    graph_set_param_val(graph, ps->clamp_d_pwmY, BOUNDS_MIN, -20000);
    graph_set_param_val(graph, ps->clamp_d_pwmY, BOUNDS_MAX, 20000);

    // TODO: Change this to use LOOP_TIME
    graph_set_param_val(graph, ps->dt, CONST_SET, 0.005);

	return 0;
 }

 int control_algorithm(log_t* log_struct, user_input_t * user_input_struct, sensor_t* sensor_struct, setpoint_t* setpoint_struct, parameter_t* parameter_struct, user_defined_t* user_defined_struct, actuator_command_t* actuator_struct, modular_structs_t* structs)
 {
    struct computation_graph* graph = parameter_struct->graph;
	// use the 'flap' switch as the flight mode selector
	int cur_fm_switch = read_flap(user_input_struct->rc_commands[FLAP]);
	static int last_fm_switch = MANUAL_FLIGHT_MODE;

	// reset flight_mode to MANUAL right away if the flap switch is in manual position
	// to engage AUTO mode the code waits for a new packet after the flap is switched to auto
	// before actually engaging AUTO mode
	if(cur_fm_switch == MANUAL_FLIGHT_MODE)
		user_defined_struct->flight_mode = MANUAL_FLIGHT_MODE;

	static float roll_trim = 0.0;
	static float pitch_trim = 0.0;

	// flap switch was just toggled to auto flight mode
	if((last_fm_switch != cur_fm_switch) && (cur_fm_switch == AUTO_FLIGHT_MODE))
	{
		user_defined_struct->engaging_auto = 1;

		// Read in trimmed values because it should read trim values right when the pilot flips the flight mode switch
		pitch_trim = user_input_struct->pitch_angle_manual_setpoint; //rc_commands[PITCH] - PITCH_CENTER;
		roll_trim = user_input_struct->roll_angle_manual_setpoint; //rc_commands[ROLL] - ROLL_CENTER;
		//sensor_struct->trimmedRCValues.yaw = yaw_manual_setpoint; //rc_commands[YAW] - YAW_CENTER;

//		sensor_struct->trims.roll = raw_actuator_struct->controller_corrected_motor_commands[ROLL];
//		sensor_struct->trims.pitch = raw_actuator_struct->controller_corrected_motor_commands[PITCH];
//		sensor_struct->trims.yaw = raw_actuator_struct->controller_corrected_motor_commands[YAW];
		sensor_struct->trims.throttle = user_input_struct->rc_commands[THROTTLE];

		log_struct->trims.roll = sensor_struct->trims.roll;
		log_struct->trims.pitch = sensor_struct->trims.pitch;
		log_struct->trims.yaw = sensor_struct->trims.yaw;
		log_struct->trims.throttle = sensor_struct->trims.throttle;
	}

	if(user_input_struct->locationFresh && user_defined_struct->engaging_auto == 1)
			user_defined_struct->engaging_auto = 2;


	// if the flap switch was toggled to AUTO_FLIGHT_MODE and we've received a new packet
	// then record the current position as the desired position
	// also reset the previous error and accumulated error from the position PIDs
	if((cur_fm_switch == AUTO_FLIGHT_MODE) && (user_defined_struct->engaging_auto == 2))
	{
        // TODO: Change this to reconnect PIDs
        /*
		// zero out the accumulated error so the I terms don't cause wild things to happen
		parameter_struct->pid_controllers[ALT_ID].acc_error = 0.0;
		parameter_struct->pid_controllers[LOCAL_X_ID].acc_error = 0.0;
		parameter_struct->pid_controllers[LOCAL_Y_ID].acc_error = 0.0;

		// make previous error equal to the current so the D term doesn't spike
		parameter_struct->pid_controllers[ALT_ID].prev_error = 0.0;
		parameter_struct->pid_controllers[LOCAL_X_ID].prev_error = 0.0;
		parameter_struct->pid_controllers[LOCAL_Y_ID].prev_error = 0.0;

		setpoint_struct->desiredQuadPosition.alt_pos = sensor_struct->currentQuadPosition.alt_pos;
		setpoint_struct->desiredQuadPosition.x_pos = sensor_struct->currentQuadPosition.x_pos;
		setpoint_struct->desiredQuadPosition.y_pos = sensor_struct->currentQuadPosition.y_pos;
		setpoint_struct->desiredQuadPosition.yaw = 0.0;//currentQuadPosition.yaw;

		// reset the flag that engages auto mode
		user_defined_struct->engaging_auto = 0;
        */
		// finally engage the AUTO_FLIGHT_MODE
		// this ensures that we've gotten a new update packet right after the switch was set to auto mode
		user_defined_struct->flight_mode = AUTO_FLIGHT_MODE;
	}

	//PIDS///////////////////////////////////////////////////////////////////////

	/* 					Position loop
	 * Reads current position, and outputs
	 * a pitch or roll for the angle loop PIDs
	 */

//		static int counter_between_packets = 0;

	if(user_input_struct->locationFresh)
	{
        // TODO: Change this to populate VRPN block
        /*
		parameter_struct->pid_controllers[LOCAL_Y_ID].current_point = sensor_struct->currentQuadPosition.y_pos;
		parameter_struct->pid_controllers[LOCAL_Y_ID].setpoint = setpoint_struct->desiredQuadPosition.y_pos;

		parameter_struct->pid_controllers[LOCAL_X_ID].current_point = sensor_struct->currentQuadPosition.x_pos;
		parameter_struct->pid_controllers[LOCAL_X_ID].setpoint = setpoint_struct->desiredQuadPosition.x_pos;

		parameter_struct->pid_controllers[ALT_ID].current_point = sensor_struct->currentQuadPosition.alt_pos;
		parameter_struct->pid_controllers[ALT_ID].setpoint = setpoint_struct->desiredQuadPosition.alt_pos;
        */
		//logging and PID computation
		//log_struct->local_y_PID_values = pid_computation(&(parameter_struct->pid_controllers[LOCAL_Y_ID]));
		//log_struct->local_x_PID_values = pid_computation(&(parameter_struct->pid_controllers[LOCAL_X_ID]));
		//log_struct->altitude_PID_values = pid_computation(&(parameter_struct->pid_controllers[ALT_ID]));

		// yaw angular position PID calculation
        graph_set_param_val(graph, parameter_struct->cur_yaw, CONST_SET, sensor_struct->currentQuadPosition.yaw);
        // TODO: Set yaw setpoint
		//parameter_struct->pid_controllers[YAW_ID].current_point = sensor_struct->currentQuadPosition.yaw;// in radians
        //parameter_struct->pid_controllers[YAW_ID].setpoint = setpoint_struct->desiredQuadPosition.yaw; // constant setpoint

		//logging and PID computation
		//log_struct->angle_yaw_PID_values = pid_computation(&(parameter_struct->pid_controllers[YAW_ID]));

	}

    graph_set_param_val(graph, parameter_struct->cur_pitch, CONST_SET, sensor_struct->pitch_angle_filtered);
    graph_set_param_val(graph, parameter_struct->cur_roll, CONST_SET, sensor_struct->roll_angle_filtered);
    graph_set_param_val(graph, parameter_struct->theta_dot, CONST_SET, sensor_struct->theta_dot);
    graph_set_param_val(graph, parameter_struct->phi_dot, CONST_SET, sensor_struct->phi_dot);
    graph_set_param_val(graph, parameter_struct->psi_dot, CONST_SET, sensor_struct->psi_dot);

    graph_set_param_val(graph, parameter_struct->rc_pitch, CONST_SET, user_input_struct->pitch_angle_manual_setpoint);
    graph_set_param_val(graph, parameter_struct->rc_roll, CONST_SET, user_input_struct->roll_angle_manual_setpoint);
    graph_set_param_val(graph, parameter_struct->rc_yaw, CONST_SET, user_input_struct->yaw_manual_setpoint);
    graph_set_param_val(graph, parameter_struct->rc_throttle, CONST_SET, user_input_struct->rc_commands[THROTTLE]);


	/*parameter_struct->pid_controllers[PITCH_ID].setpoint =
			(user_defined_struct->flight_mode == AUTO_FLIGHT_MODE)?
			(parameter_struct->pid_controllers[LOCAL_X_ID].pid_correction) + pitch_trim : user_input_struct->pitch_angle_manual_setpoint;

	parameter_struct->pid_controllers[ROLL_ID].setpoint =
			(user_defined_struct->flight_mode == AUTO_FLIGHT_MODE)?
			(parameter_struct->pid_controllers[LOCAL_Y_ID].pid_correction) + roll_trim : user_input_struct->roll_angle_manual_setpoint;
    */
    /*
    int outputs[3] = {parameter_struct->pitch_r_pid,
                             parameter_struct->roll_r_pid,
                             parameter_struct->yaw_r_pid};*/
    int outputs[1] = {parameter_struct->mixer};
    graph_compute_nodes(graph, outputs, 1);

	 // here for now so in case any flight command is not PID controlled, it will default to rc_command value:
	//memcpy(raw_actuator_struct->controller_corrected_motor_commands, user_input_struct->rc_commands, sizeof(int) * 6);

	// don't use the PID corrections if the throttle is less than about 10% of its range
	if((user_input_struct->rc_commands[THROTTLE] >
	118000) || (user_defined_struct->flight_mode == AUTO_FLIGHT_MODE))
	{

		if(user_defined_struct->flight_mode == AUTO_FLIGHT_MODE)
		{
            /*
			//THROTTLE
			raw_actuator_struct->controller_corrected_motor_commands[THROTTLE] =
				((int)(parameter_struct->pid_controllers[ALT_ID].pid_correction)) + sensor_struct->trims.throttle;

			//ROLL
			raw_actuator_struct->controller_corrected_motor_commands[ROLL] =
					parameter_struct->pid_controllers[ROLL_RATE_ID].pid_correction; // + sensor_struct->trims.roll;

			//PITCH
			raw_actuator_struct->controller_corrected_motor_commands[PITCH] =
					parameter_struct->pid_controllers[PITCH_RATE_ID].pid_correction; // + sensor_struct->trims.pitch;

			//YAW
			raw_actuator_struct->controller_corrected_motor_commands[YAW] =
					parameter_struct->pid_controllers[YAW_RATE_ID].pid_correction;// + sensor_struct->trims.yaw;
            */

//			static int slow_down = 0;
//			slow_down++;
//			if(slow_down % 50 == 0)
//			printf("X: %.3f\tY: %.3f\tZ: %.3f\tX_s: %.3f\tX_c: %.3f\tY_s: %.3f\tY_c: %.3f\tZ_s: %.3f\tZ_c: %.3f\t\n",
//					parameter_struct->pid_controllers[LOCAL_X_ID].pid_correction,
//					parameter_struct->pid_controllers[LOCAL_Y_ID].pid_correction,
//					parameter_struct->pid_controllers[ALT_ID].pid_correction,
//					parameter_struct->pid_controllers[LOCAL_X_ID].setpoint, parameter_struct->pid_controllers[LOCAL_X_ID].current_point,
//					parameter_struct->pid_controllers[LOCAL_Y_ID].setpoint, parameter_struct->pid_controllers[LOCAL_Y_ID].current_point,
//					parameter_struct->pid_controllers[ALT_ID].setpoint, parameter_struct->pid_controllers[ALT_ID].current_point);
		}
		else{
			//THROTTLE
			actuator_struct->pwms[0] = graph_get_output(graph, parameter_struct->mixer, MIXER_PWM0);

			actuator_struct->pwms[1] = graph_get_output(graph, parameter_struct->mixer, MIXER_PWM1);

			actuator_struct->pwms[2] = graph_get_output(graph, parameter_struct->mixer, MIXER_PWM2);

			actuator_struct->pwms[3] = graph_get_output(graph, parameter_struct->mixer, MIXER_PWM3);
		}

        /*
		//BOUNDS CHECKING
		if(raw_actuator_struct->controller_corrected_motor_commands[THROTTLE] < 0)
			raw_actuator_struct->controller_corrected_motor_commands[THROTTLE] = 0;

		//BOUNDS CHECKING
		if(raw_actuator_struct->controller_corrected_motor_commands[ROLL] > 20000)
			raw_actuator_struct->controller_corrected_motor_commands[ROLL] = 20000;

		if(raw_actuator_struct->controller_corrected_motor_commands[ROLL] < -20000)
			raw_actuator_struct->controller_corrected_motor_commands[ROLL] = -20000;

		if(raw_actuator_struct->controller_corrected_motor_commands[PITCH] > 20000)
			raw_actuator_struct->controller_corrected_motor_commands[PITCH] = 20000;

		if(raw_actuator_struct->controller_corrected_motor_commands[PITCH] < -20000)
			raw_actuator_struct->controller_corrected_motor_commands[PITCH] = -20000;

		if(raw_actuator_struct->controller_corrected_motor_commands[YAW] > 20000)
			raw_actuator_struct->controller_corrected_motor_commands[YAW] = 20000;

		if(raw_actuator_struct->controller_corrected_motor_commands[YAW] < -20000)
			raw_actuator_struct->controller_corrected_motor_commands[YAW] = -20000;
		*/
	}
	else
	{
		actuator_struct->pwms[0] = user_input_struct->rc_commands[THROTTLE];
		actuator_struct->pwms[1] = user_input_struct->rc_commands[THROTTLE];
		actuator_struct->pwms[2] = user_input_struct->rc_commands[THROTTLE];
		actuator_struct->pwms[3] = user_input_struct->rc_commands[THROTTLE];
		/*
		raw_actuator_struct->controller_corrected_motor_commands[ROLL] = 0;
		raw_actuator_struct->controller_corrected_motor_commands[PITCH] = 0;
		raw_actuator_struct->controller_corrected_motor_commands[YAW] = 0;*/
	}

	//logging
	// here we are not actually duplicating the logging from the PID computation
	// the PID computation logs PID_values struct where this logs the PID struct
	// they contain different sets of data
    /*
	log_struct->local_y_PID = parameter_struct->pid_controllers[LOCAL_Y_ID];
	log_struct->local_x_PID = parameter_struct->pid_controllers[LOCAL_X_ID];
	log_struct->altitude_PID = parameter_struct->pid_controllers[ALT_ID];

	log_struct->angle_roll_PID = parameter_struct->pid_controllers[ROLL_ID];
	log_struct->angle_pitch_PID = parameter_struct->pid_controllers[PITCH_ID];
	log_struct->angle_yaw_PID = parameter_struct->pid_controllers[YAW_ID];

	log_struct->ang_vel_roll_PID = parameter_struct->pid_controllers[ROLL_RATE_ID];
	log_struct->ang_vel_pitch_PID = parameter_struct->pid_controllers[PITCH_RATE_ID];
	log_struct->ang_vel_yaw_PID = parameter_struct->pid_controllers[YAW_RATE_ID];
    */
	last_fm_switch = cur_fm_switch;

	// Make location stale now
	user_input_struct->locationFresh = 0;

    return 0;
 }

 void setPIDCoeff(PID_t* p, float pValue, float iValue, float dValue) {

 	p->Kp = pValue;
 	p->Ki = iValue;
 	p->Kd = dValue;

 }
